/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.integration;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.EnumSet;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.net.ssl.SSLContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.config.SocketConfig;
import org.apache.http.conn.HttpClientConnectionManager;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLContexts;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;

import cz.tdp.kshield.client.ClientMessage;
import cz.tdp.kshield.client.KShieldClient;
import cz.tdp.kshield.client.KShieldClientException;
import cz.tdp.kshield.client.ResponseLevel;
import cz.tdp.kshield.client.UserInfo;
import cz.tdp.kshield.client.UserInfo.AuthType;

/**
 * Basic implementation of KeyShield SSO Server AuthenticationService.
 * KShieldClient is used internally to communicate with KeyShield SSO Server.
 * 
 * @see cz.tdp.kshield.client.KShieldClient
 */
public class SimpleAuthenticationServiceImpl implements AuthenticationService
{
  protected KShieldClient client;

  protected HttpClientConnectionManager connManager;
  protected CloseableHttpClient httpClient;
  
  /**
   * @param url KeyShield SSO Server url
   */
  public SimpleAuthenticationServiceImpl(String url) {
    this.url = url;
  }

  protected void checkUrl() {
    if (url == null || url.length() == 0) throw new IllegalArgumentException("Please provide valid KeyShield SSO server URL");
  }

  /**
   * Initializes Authentication service after creation (In Spring v 3.x and Guice IOC called automatically)
   * 
   * <p><em>Important - call this method after creation and overall setup of AuthenticationService instance
   */
  @Override
  @PostConstruct
  public void init() {
    log.info("Starting KeyShield SSO AuthenticationService");
    
    checkUrl();
    
    if (httpClient == null) {
      log.info("Creating HttpClientConnectionManager");
      Registry<ConnectionSocketFactory> registry = RegistryBuilder.<ConnectionSocketFactory>create()
          .register("http", PlainConnectionSocketFactory.getSocketFactory())
          .register("https", createSSLSocketFactory())
          .build();
      
      connManager = new PoolingHttpClientConnectionManager(registry);
      
      log.info("Creating HttpClient");
      SocketConfig socketConfig = SocketConfig.custom()
          .setTcpNoDelay(true)
          .build();
      
      RequestConfig requestConfig = RequestConfig.custom()
          .setSocketTimeout(soTimeout)
          .setConnectTimeout(connectionTimeout)
          .build();
      
      httpClient = HttpClients.custom()
          .setConnectionManager(connManager)
          .setDefaultSocketConfig(socketConfig)
          .setDefaultRequestConfig(requestConfig)
          .build();
    }
    
    log.info("Create KShieldClient: kshieldUrl=" + url + ", apiKey=" + apiKey);
    
    if (hasOptionalAttributes()) {
      client = new KShieldClient(url, apiKey, httpClient, getMergedAttributes());
    }
    else {
      client = new KShieldClient(url, apiKey, httpClient);
    }
    
    log.info("Setting Allowed Auth types to: " + allowedAuthTypes);
    client.setAllowedAuthTypes(allowedAuthTypes);
    
    log.info("Setting ResponseLevel to: " + responseLevel);
    client.setResponseLevel(responseLevel);
  }

  /**
   * Cleanup Authentication service before destruction (In Spring v 3.x and Guice IOC called automatically)
   */
  @Override
  @PreDestroy
  public void destroy() {
    log.info("Clean up KeyShield SSO AuthenticationService");
    
    if (httpClient != null) {
      try {
        httpClient.close();
      }
      catch (IOException e) {
        log.warn("error while closing httpClient", e);
      }
      
      httpClient = null;
    }
    
    if (connManager != null) {
      connManager.shutdown();
      connManager = null;
    }
    
    client = null;
    
    log.info("Clean up finished");
  }

  @Override
  public void checkService() throws KShieldClientException {
    if (client == null) throw new IllegalStateException("This AuthenticaticationService must be initialized (method init) before attempt to authenticate");
    
    client.getUserByIP(TEST_IP_ADDR);
  }
  
  /**
   * @param ipAddr remote request IP address
   * @return new instance of <code>UserInfo</code> if successfully authenticated or null otherwise
   */
  @Override
  public UserInfo createUserInfo(String ipAddr) {
    if (client == null) throw new IllegalStateException("This AuthenticaticationService must be initialized (method init) before attempt to authenticate");
    
    if (ipAddr == null || ipAddr.length() == 0) {
      //XXX maybe throw IllegalArgumentException?
      return null;
    }
    
    try {
      KShieldContext.initKShieldSession();
      
      final UserInfo userInfo = client.getUserByIP(ipAddr, usernameAttribute);
      
      if (userInfo != null && userInfoValidator.validate(userInfo)) {
        // kshield session is closed by caller of this method
        
        KShieldContext.startKShieldSession();
        return userInfo;
      }
    }
    catch (KShieldClientException e) {
      log.error("client error", e);
    }
    
    KShieldContext.closeKShieldSession();
    return null;
  }
  
  @Override
  public void sendClientMessage(String from, String to, String message) throws KShieldClientException {
    sendClientMessage(new ClientMessage(from, to, message));
  }
  
  @Override
  public void sendClientMessage(ClientMessage msg) throws KShieldClientException {
    if (client == null) throw new IllegalStateException("This AuthenticaticationService must be initialized (method init) before attempt to authenticate");
    
    try {
      client.sendClientMessage(msg);
    }
    catch (KShieldClientException e) {
      log.error("client error", e);
      throw e;
    }
  }
  
  protected String url;
  
  /**
   * @return KeyShield SSO Server url
   */
  public String getUrl() {
    return this.url;
  }

  /**
   * Sets KeyShield SSO Server url
   * <p><em>Important - set this before init() method call
   */
  public void setUrl(String url) {
    this.url = url;
  }
  
  private String apiKey;
  
  /**
   * @return KeyShield SSO API authorization key
   */
  public String getApiKey() {
    return this.apiKey;
  }

  /**
   * Sets KeyShield SSO API authorization key
   * <p><em>Important - set this before init() method call
   * @param apiKey
   */
  public void setApiKey(String apiKey) {
    this.apiKey = apiKey;
  }
  
  private String usernameAttribute;
  
  /**
   * @return UserInfo attribute used as username
   */
  public String getUsernameAttribute() {
    return usernameAttribute;
  }
  
  /**
   * Set name of attribute used as username instead of screenName
   * This attribute is automatically merged with optional attributes
   * <p><em>Important - set this before init() method call
   * 
   * @param usernameAttr name of username attribute
   */
  public void setUsernameAttribute(String usernameAttr) {
    this.usernameAttribute = usernameAttr;
  }

  private static final String[] EMPTY_ATTRS = new String[0];
  
  private String[] optionalAttributes = EMPTY_ATTRS;
  
  /**
   * Set optional attributes requested from KeyShield SSO with UserInfo
   * <p><em>Important - set this before init() method call
   * 
   * @param attrs optional attributes names
   */
  public void setOptionalAttributes(String... attrs) {
    if (attrs != null && attrs.length > 0) {
      this.optionalAttributes = attrs;
    }
    else {
      this.optionalAttributes = EMPTY_ATTRS;
    }
  }

  protected String[] getMergedAttributes() {
    final boolean hasUserAttr = hasUsernameAttr();
    final int optAttrSize = optionalAttributes.length;
    
    if (hasUserAttr || optAttrSize > 0) {
      final String[] merged = new String[optAttrSize + (hasUserAttr ? 1 : 0)];
      
      if (hasUserAttr) {
        merged[0] = usernameAttribute;
      }
      if (optAttrSize > 0) {
        final int optAttrIndex = hasUserAttr ? 1 : 0;
        System.arraycopy(optionalAttributes, 0, merged, optAttrIndex, optAttrSize);
      }
      
      return merged;
    }
    
    return null;
  }

  private EnumSet<AuthType> allowedAuthTypes = UserInfo.DEFAULT_ALLOWED_AUTH_TYPES;

  /**
   * @returns allowed authentication types
   * @see AuthType
   */
  public EnumSet<AuthType> getAllowedAuthTypes() {
    return allowedAuthTypes;
  }

  /**
   * Sets allowed authentication types
   * <p><em>Important - set this before init() method call
   * 
   * @param allowAuthTypes set of allowed authentication types
   * @see AuthType
   */
  public void setAllowedAuthTypes(EnumSet<AuthType> allowAuthTypes) {
    this.allowedAuthTypes = allowAuthTypes;
  }
  
  private ResponseLevel responseLevel;

  /**
   * @return configured client response level
   * @see ResponseLevel
   */
  public ResponseLevel getResponseLevel() {
    return this.responseLevel;
  }

  /**
   * Sets optional response level used in KeyShieldSSO requests
   * It is possible to set this dynamically after init() method
   * 
   * @param responseLevel
   */
  public void setResponseLevel(ResponseLevel responseLevel) {
    this.responseLevel = responseLevel;
    
    if (client != null) {
      client.setResponseLevel(responseLevel);
    }
  }

  /**
   * Connection timeout
   */
  private int connectionTimeout = 5000;
  
  /**
   * Returns http connection timeout in milliseconds - default is 5000
   */
  public int getConnectionTimeout() {
    return connectionTimeout;
  }

  /**
   * Sets http connection timeout in milliseconds - default is 5000
   * <p><em>Important - set this before init() method call
   * 
   * @param connectionTimeout
   */
  public void setConnectionTimeout(int connectionTimeout) {
    this.connectionTimeout = connectionTimeout;
  }
  
  /**
   * Socket timeout
   */
  private int soTimeout = 5000;
  
  /**
   * Return SO_TIMEOUT in milliseconds - default is 5000
   */
  public int getSoTimeout() {
    return soTimeout;
  }

  /**
   * Set SO_TIMEOUT in milliseconds - default is 5000
   * <p><em>Important - set this before init() method call
   * 
   * @param soTimeout
   */
  public void setSoTimeout(int soTimeout) {
    this.soTimeout = soTimeout;
  }
  
  private UserInfoValidator userInfoValidator = new SimpleUserInfoValidator();

  /**
   * Return custom userInfo validator
   */
  public UserInfoValidator getUserInfoValidator() {
    return this.userInfoValidator;
  }

  /**
   * Set custom userInfo validator
   */
  public void setUserInfoValidator(UserInfoValidator userInfoValidator) {
    if (userInfoValidator != null) {
      this.userInfoValidator = userInfoValidator;
    }
    else {
      this.userInfoValidator = new SimpleUserInfoValidator();
    }
  }
  
  private String trustStorePath;
  
  /**
   * @return Custom JKS truststore path
   */
  public String getTrustStorePath() {
    return this.trustStorePath;
  }

  /**
   * Sets custom JKS truststore path
   * This truststore will be used as in memory keystore - all certificates are treated as trusted
   * <p><em>Important - set this before init() method call
   * 
   * @param trustStorePath
   */
  public void setTrustStorePath(String trustStorePath) {
    this.trustStorePath = trustStorePath;
  }

  protected boolean hasOptionalAttributes() {
    return hasUsernameAttr() || optionalAttributes.length > 0;
  }

  protected boolean hasUsernameAttr() {
    return usernameAttribute != null && !usernameAttribute.isEmpty();
  }
  
  protected boolean hasTrustStore() {
    return trustStorePath != null && !trustStorePath.isEmpty();
  }
  
  protected SSLConnectionSocketFactory createSSLSocketFactory() {
    try {
      if (hasTrustStore()) {
        log.info("Loading Truststore from: " + trustStorePath);
        
        final KeyStore truststore = loadTrustStore();
        
        if (truststore != null) {
          log.info("Create InMemoryTrustStoreSSLFactory");
          
          final SSLContext sslContext = SSLContexts.custom()
              .loadTrustMaterial(truststore)
              .build();
          
          return new SSLConnectionSocketFactory(sslContext);
        }
      }
    }
    catch (KeyManagementException|NoSuchAlgorithmException|KeyStoreException e) {
      log.error("failed create SSL Socket Factory", e);
    }
    
    return SSLConnectionSocketFactory.getSocketFactory();
  }
  
  protected KeyStore loadTrustStore() {
    try (InputStream in = new FileInputStream(trustStorePath)) {
      KeyStore trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
      trustStore.load(in, null);
      return trustStore;
    }
    catch (NoSuchAlgorithmException|CertificateException|KeyStoreException|IOException e) {
      log.error("Cannot load Trust Store", e);
    }
    
    log.warn("Unable to load truststore from path: " + trustStorePath);
    return null;
  }
  
  //NOSONAR-BEGIN hard coded IP
  private static final String TEST_IP_ADDR = "0.0.0.0";
  //NOSONAR-END

  private static final Log log = LogFactory.getLog(SimpleAuthenticationServiceImpl.class);
}
