/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.integration.web;

import java.util.List;

import javax.servlet.http.HttpServletRequest;

import cz.tdp.kshield.client.ClientMessage;
import cz.tdp.kshield.client.KShieldClientException;
import cz.tdp.kshield.client.UserInfo;
import cz.tdp.kshield.integration.AuthenticationService;
import cz.tdp.kshield.integration.ServletAuthenticationService;

/**
 * Implementation of KeyShield SSO Server AuthenticationService that delegates all functionality to list of AuthenticationService instances.
 */
public class DelegatingAuthenticationServiceImpl implements ServletAuthenticationService
{
  /**
   * list of AuthenticationService instances
   */
  private final List<ServletAuthenticationService> services;
  
  /**
   * @param services list of AuthenticationService instances
   */
  public DelegatingAuthenticationServiceImpl(List<ServletAuthenticationService> services) {
    if (services == null || services.isEmpty()) throw new IllegalArgumentException("Please provide valid list of KeyShield SSO AuthenticationService instances");
    
    this.services = services;
  }
  
  @Override
  public void init() {
    for (AuthenticationService service : services) {
      service.init();
    }
  }
  
  @Override
  public void destroy() {
    for (AuthenticationService service : services) {
      service.destroy();
    }
  }
  
  @Override
  public void checkService() throws KShieldClientException {
    for (AuthenticationService service : services) {
      service.checkService();
    }
  }
  
  /**
   * Return true if any service returns true
   */
  @Override
  public boolean authenticate(HttpServletRequest request) {
    for (ServletAuthenticationService service : services) {
      if (service.authenticate(request)) {
        return true;
      }
    }
    
    return false;
  }

  /**
   * Return first non null userInfo
   * 
   * @param ipAddr remote request IP address
   */
  @Override
  public UserInfo createUserInfo(String ipAddr) {
    for (AuthenticationService service : services) {
      final UserInfo userInfo = service.createUserInfo(ipAddr);
      
      if (userInfo != null) {
        return userInfo;
      }
    }
    
    return null;
  }
  
  @Override
  public void sendClientMessage(String from, String to, String message) throws KShieldClientException {
    sendClientMessage(new ClientMessage(from, to, message));
  }
  
  @Override
  public void sendClientMessage(ClientMessage msg) throws KShieldClientException {
    KShieldClientException lastEx = null;
    
    for (AuthenticationService service : services) {
      try {
        service.sendClientMessage(msg);
      }
      catch (KShieldClientException e) {
        // exception logged
        lastEx = e;
      }
    }
    
    if (lastEx != null) {
      //XXX maybe aggregate all errors and throw some other exception something like KShieldMutipleErrorsException
      throw lastEx;
    }
  }
}
