/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.integration.web;

import java.io.IOException;
import java.lang.reflect.Constructor;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.BeanWrapper;
import org.springframework.beans.PropertyAccessorFactory;

import cz.tdp.kshield.integration.KShieldContext;
import cz.tdp.kshield.integration.ServletAuthenticationService;


/**
 * <b>This is main class of web intergation package.</b><br/>
 * <br/>
 * This filter could be used in pure Servlet 2.5 environment and via Spring Framework. You need to position this filter before any authentication check in your application.
 * Retrieved UserInfo instance is stored in session attribute {@link ServletAuthenticationService#KSHIELD_PRINCIPAL_TOKEN} after successful validation.
 * You can use this data for furher validation or other authenticanion steps required in your application.<br/>
 * <b>Example configuration in web.xml:</b>
<pre>{@code

<filter>
  <filter-name>KShieldAuthFilter</filter-name>
  <filter-class>cz.tdp.kshield.integration.web.KShieldAuthenticationFilter</filter-class>
  <init-param>
    <param-name>serviceClass</param-name>
    <param-value>cz.tdp.kshield.integration.DefaultAuthenticationServiceImpl</param-value>
  </init-param>
  <init-param>
    <param-name>serverUrl</param-name>
    <param-value>http://127.0.0.1:8485</param-value>
  </init-param>
</filter>

<filter-mapping>
  <filter-name>KShieldAuthFilter</filter-name>
  <url-pattern>*.jsp</url-pattern>
</filter-mapping>

}</pre>
 *
 * @see #init(FilterConfig)
 * @see DefaultAuthenticationServiceImpl
 * @see Filter
 */
public class KShieldAuthenticationFilter implements Filter
{
  @Override
  public void doFilter(ServletRequest request, ServletResponse response, FilterChain filters) throws IOException, ServletException {
    if (request instanceof HttpServletRequest) {
      final HttpServletRequest httpReq = (HttpServletRequest)request;
      
      if (authenticationService.authenticate(httpReq)) {
        log.info("Authenticated to KeyShield SSO Server");
      }
    }
    
    try {
      filters.doFilter(request, response);
    }
    finally {
      KShieldContext.closeKShieldSession();
    }
  }

  @Override
  public void init(FilterConfig config) throws ServletException {
    if (authenticationService == null) {
      // initialize sso authentication service from filter init parameters
      final String serverUrl = config.getInitParameter("serverUrl");
      
      if (serverUrl != null) {
        final String serviceClassName = config.getInitParameter("serviceClass");
        
        if (serviceClassName != null) {
          try {
            @SuppressWarnings("unchecked")
            final Class<ServletAuthenticationService> serviceClass = (Class<ServletAuthenticationService>)Class.forName(serviceClassName);
            final Constructor<ServletAuthenticationService> constructor = serviceClass.getConstructor(String.class);
            
            authenticationService = constructor.newInstance(serverUrl);
          }
          catch (Exception e) {
            log.error("Cannot create authentication service", e);
            throw new ServletException("Failed create kshield authentication service instance", e);
          }
        }
        else {
          authenticationService = new DefaultAuthenticationServiceImpl(serverUrl);
        }
        
        final BeanWrapper wrapper = PropertyAccessorFactory.forBeanPropertyAccess(authenticationService);
        
        final String apiKey = config.getInitParameter("apiKey");
        if (apiKey != null) {
          wrapper.setPropertyValue("apiKey", apiKey);
        }
        
        final String usernameAttr = config.getInitParameter("usernameAttr");
        if (usernameAttr != null) {
          wrapper.setPropertyValue("usernameAttribute", usernameAttr);
        }
        
        authenticationService.init();
      }
      else {
        throw new ServletException("Missing serverUrl init parameter");
      }
    }
  }

  @Override
  public void destroy() {
    //do nothing
  }

  private ServletAuthenticationService authenticationService;
  
  public void setAuthenticationService(ServletAuthenticationService authService) {
    this.authenticationService = authService;
  }

  private static final Log log = LogFactory.getLog(KShieldAuthenticationFilter.class);
}
