/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package cz.tdp.kshield.notification;

import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Common Abstract WebSocket Endpoint class
 */
public abstract class AbstractWebSocketEndpoint implements KShieldEndpoint
{
  private static final Log log = LogFactory.getLog(AbstractWebSocketEndpoint.class);
  
  private static final AtomicInteger counter = new AtomicInteger(1);
  private final int wsId = counter.getAndIncrement();
  
  private final AtomicBoolean initialized = new AtomicBoolean(false);
  
  private String ipAddress;
  private String user;
  private String kshieldAuth;
  private String lastError;
  
  /**
   * Implement this to provide kshieldAuthManager functionality to WebSocket instance
   * @return reference to kshieldAuthManager
   */
  protected abstract KShieldAuthManager kshieldAuthManager();
  
  public AbstractWebSocketEndpoint() {
    super();
  }

  @Override
  public boolean isActive() {
    return ipAddress != null;
  }
  
  @Override
  public void signal() {
    if (isActive()) {
      send(poll());
    }
  }
  
  @Override
  public String toString() {
    return "user="+user+", kshield="+kshieldAuth+", id=" + wsId + (lastError!=null?" ["+lastError+"]":"");
  }
  
  protected void onErrorImpl(Throwable e) {
    onCloseImpl();
    
    lastError = e.getMessage();
    
    log.trace("WebSocket error: ", e);
  }
  
  protected String onMessageImpl(String message) {
    if (message.startsWith("POLL")) {
      if (log.isDebugEnabled()) {
        log.debug("WebSocket [id="+wsId+"] message: " + message);
      }
      
      handleMessage(message);
      
      if (isActive()) {
        if (!initialized.getAndSet(true)) {
          kshieldAuthManager().addEndpoint(ipAddress, this);
        }
      
        return poll();
      }
    }
    
    // echo - for testing purposes
    return message;
  }
  
  private void handleMessage(String message) {
    if (!initialized.get()) {
      final String[] tokens = message.split("[:]");
      
      if (tokens.length > 1) {
        ipAddress = token(tokens, 1);
        user = token(tokens, 2);
        kshieldAuth = token(tokens, 3);
      }
    }
  }
  
  protected abstract void send(String message);
  
  protected void onOpenImpl() {
    reset();
  }
  
  protected void onCloseImpl() {
    if (ipAddress != null) {
      kshieldAuthManager().removeEndpoint(ipAddress, this);
    }
    
    reset();
  }

  protected void reset() {
    ipAddress = null;
    user = null;
    kshieldAuth = null;
    lastError = null;
    
    initialized.set(false);
  }

  private String poll() {
    return kshieldAuthManager().poll(ipAddress, kshieldAuth, isNotEmpty(user));
  }

  private String token(String[] tokens, int idx) {
    if (idx < tokens.length) {
      return trimToNull(tokens[idx]);
    }
    
    return null;
  }

  private String trimToNull(String str) {
    if (isEmpty(str) || "null".equals(str)) {
      return null;
    }
    return str.trim();
  }
  
  private boolean isEmpty(String str) {
    return str == null || str.trim().isEmpty();
  }
  
  protected boolean isNotEmpty(String str) {
    return str != null && !str.isEmpty();
  }
}
