/*
 * Copyright (c) 2013 - 2016 TDP Ltd All Rights Reserved.
 * TDP Ltd grants permission, free of charge, to any person obtaining copies
 * of this software and its associated documentation files (the "Software"),
 * to deal in the Software without restriction, including to use, copy, adapt,
 * publish, distribute, display, perform, sublicense, and sell copies of the
 * Software, subject to the following condition: You must include the above
 * copyright notice and this permission notice in all full or partial copies
 * of the Software.
 * 
 * TDP LTD PROVIDES THE SOFTWARE "AS IS," WITHOUT ANY EXPRESS OR IMPLIED WARRANTY,
 * INCLUDING WITHOUT THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGMENT. TDP LTD, THE AUTHORS OF THE SOFTWARE,
 * AND THE OWNERS OF COPYRIGHT IN THE SOFTWARE ARE NOT LIABLE FOR ANY CLAIM, DAMAGES,
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT, OR OTHERWISE, ARISING
 * FROM, OUT OF, OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.tdp.kshield.springsecurity3;


import java.io.IOException;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.web.authentication.preauth.AbstractPreAuthenticatedProcessingFilter;

import cz.tdp.kshield.client.UserInfo;
import cz.tdp.kshield.integration.AuthenticationService;
import cz.tdp.kshield.integration.KShieldContext;

/**
 * <b>This is central class in springsecurity3 package.</b> You need to define AuthenticationService and AuthenticationFactory beans (usually from package springsecurity3).
 * You can use your own AuthenticationFactory implementation or extend AuthenticationService if you need do some post or pre processing during authentication process.<br/>
 * Default implementations depends also on UserDetailsService and AuthenticationDetailsSource.<br/>
 * <br/>
 * <b>Example spring configuration:</b>
<pre>{@code

<bean id="userDetailsService" class="org.springframework.security.core.userdetails.memory.InMemoryDaoImpl">
 <property name="userProperties">
    <props>
       <prop key="user1">password</prop>
    </props>
  </property>
</bean>
 
<bean id="kshieldAuthenticationFactory" class="cz.tdp.kshield.springsecurity3.SpringAuthenticationFactory" autowire="byType" />
 
<bean id="kshieldAuthService" class="cz.tdp.kshield.springsecurity3.SpringAuthenticationServiceImpl" autowire="byType">
  <constructor-arg value="http://127.0.0.1:8485" />
</bean>
 
<bean id="authenticationDetailsSource" class="org.springframework.security.core.userdetails.UserDetailsByNameServiceWrapper" autowire="byType" />

<bean id="kshieldAuthenticationFilter" class="cz.tdp.kshield.springsecurity3.KShieldPreAuthenticationFilter" autowire="byType">
  <property name="authenticationDetailsSource" ref="authenticationDetailsSource" />
  <property name="authenticationManager" ref="authenticationManager" />
  <property name="authenticationService" ref="kshieldAuthService" />
</bean>

}</pre>
 * 
 * <b>You should add PreAuthenticatedAuthenticationProvider provider to authentication manager configuration and define alias:</b>
<pre>{@code

<bean id="preAuthenticatedAuthenticationProvider" class="org.springframework.security.web.authentication.preauth.PreAuthenticatedAuthenticationProvider" autowire="byType" />
 
<security:authentication-manager alias="authenticatedManager">
  <security:authentication-provider ref="preAuthenticatedAuthenticationProvider" />
  ...
</security:authentication-manager>

}</pre>
 * 
 * <b>Finally you need add custom filter (defined above) to your http element config:</b>
<pre>{@code

<security:http>
  ...
  <security:custom-filter position="PRE_AUTH_FILTER" ref="kshieldAuthenticationFilter" />
</security:http>

}</pre>
 * @see AbstractPreAuthenticatedProcessingFilter
 */
public class KShieldPreAuthenticationFilter extends AbstractPreAuthenticatedProcessingFilter
{
  /**
   * Suffix used in fake use password
   */
  public static final String KSHIELD_PASSWD_SUFFIX = "\034\035\036\037kshield";
  
  public KShieldPreAuthenticationFilter() {
    super();
    
    this.setCheckForPrincipalChanges(false);
  }
  
  @Override
  public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
    try {
      super.doFilter(request, response, chain);
    }
    finally {
      KShieldContext.closeKShieldSession();
    }
  }
  
  @Override
	protected Object getPreAuthenticatedCredentials(HttpServletRequest request) {
		return request.getRemoteAddr()+KSHIELD_PASSWD_SUFFIX;
	}

	@Override
	protected Object getPreAuthenticatedPrincipal(HttpServletRequest request) {
	  final UserInfo info = authenticationService.createUserInfo(request.getRemoteAddr());
	  
	  if (info != null) {
	    return getPrincipal(info);
	  }
	  
    return null;
	}

	/**
   * @param userInfo UserInfo class returned by KeyShield SSO Server client
   * @return username retrieved from UserInfo
   */
  protected String getPrincipal(UserInfo userInfo) {
    return userInfo.getUsername();
  }
	
	@Autowired
	private AuthenticationService authenticationService;
	
	public void setAuthenticationService(AuthenticationService authService) {
    this.authenticationService = authService;
  }

  protected final Log log = LogFactory.getLog(getClass());
}
